package dev.pfaff.altargui;

import dev.pfaff.altargui.handlers.*;
import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.internal.elements.RootElement;
import dev.pfaff.altargui.internal.render.ScissorStack;
import dev.pfaff.altargui.pathfinder.PathfinderDemo;
import dev.pfaff.altargui.util.Hasher;
import dev.pfaff.altargui.util.NonNullLong2WidgetInstanceHashMap;
import dev.pfaff.altargui.util.listenable.EventListenable;
import dev.pfaff.log4truth.Logger;
import it.unimi.dsi.fastutil.longs.LongArrayList;
import it.unimi.dsi.fastutil.shorts.ShortArrayList;
import org.jspecify.annotations.Nullable;

import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import java.util.function.Predicate;

import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.GuiGlobals.profiler;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static java.util.FormatProcessor.FMT;

public final class GuiContext {
	private static final RootElement USER_ROOT = new RootElement(0);
	static final RootElement INSPECTOR_ROOT = new RootElement("INSPECTO");

	private final NonNullLong2WidgetInstanceHashMap elements = new NonNullLong2WidgetInstanceHashMap(64);
	private final LongArrayList elementsRemovalQueue = new LongArrayList();
	private final WidgetInstanceList orderedBuffer = new WidgetInstanceList(512);
	private final ShortArrayList treeBuffer = new ShortArrayList();

	private int scaledWindowWidth;
	private int scaledWindowHeight;

	private final CanvasImpl canvas = new CanvasImpl();
	private final MountContextImpl mountContext = new MountContextImpl(this);
	private final StateContextImpl stateContext = new StateContextImpl();

	public final EventListenable preUpdate = new EventListenable();

	private final Widget widget;
	@Nullable
	private WidgetInstance<?> widgetElement;
	private final Inspector inspector = new Inspector(this, orderedBuffer, treeBuffer, canvas.enabledDebugLayers, canvas);
	@Nullable
	private WidgetInstance<?> inspectorElement;
	private float mouseX;
	private float mouseY;

	public GuiContext(Widget widget) {
		this.widget = widget;
	}

	public void dispose() {
		elements.forEachValue(WidgetInstance::dispose);
		canvas.close();
	}

	/**
	 * @param scaleFactor the scale factor that should be applied to the UI.
	 * @param windowWidth the framebuffer width, unscaled.
	 * @param windowHeight the framebuffer height, unscaled.
	 */
	public void setWindowProps(double scaleFactor, int windowWidth, int windowHeight) {
		this.scaledWindowWidth = (int)Math.round(windowWidth / scaleFactor);
		this.scaledWindowHeight = (int)Math.round(windowHeight / scaleFactor);
		Logger.log(() -> STR."setWindowProps: scaleFactor=\{scaleFactor}, windowWidth=\{windowWidth}, windowHeight=\{windowHeight} -> scaledWindowWidth=\{scaledWindowWidth}, scaledWindowHeight=\{scaledWindowHeight}");
		canvas.setWindowProps(scaleFactor, windowWidth, windowHeight);
	}

	public void updateAndPaint() {
		profiler().push("GuiContext.updateAndPaint");
		this.mountContext.time = System.currentTimeMillis();
		var canvas = this.canvas;
		orderedBuffer.clear();
		treeBuffer.clear();
		canvas.reset();

		preUpdate.notifyListeners();

		{
			profiler().push("build user content");
			var element = widgetElement;
			if (element == null) element = widgetElement = stateContext.getState(USER_ROOT.key(), () -> {}, widget);
			profiler().replace("lay out user content");
			layoutRootElement(USER_ROOT, element);
			profiler().replace("mount user content");
			mountContext.mount(element);
			profiler().pop();
		}

		{
			inspector.inspectableCount = orderedBuffer.size();
			profiler().push("build inspector content");
			var element = inspectorElement;
			if (element == null) element = inspectorElement = stateContext.getState(INSPECTOR_ROOT.key(), () -> {}, inspector);
			profiler().replace("lay out inspector content");
			layoutRootElement(INSPECTOR_ROOT, element);
			profiler().replace("mount inspector content");
			mountContext.mount(element);
			profiler().pop();
		}

		profiler().push("Canvas.draw");
		canvas.draw();
		profiler().pop();

		boolean generation = this.mountContext.generation;
		// prune the element state
		{
			var elementsRemovalQueue = this.elementsRemovalQueue;
			var elements = this.elements;
			elements.forEachValue(element -> {
				if (element.generation != generation) {
					element.dispose();
					elementsRemovalQueue.add(element.key);
				}
			});
			elementsRemovalQueue.forEach(elements::remove);
			elementsRemovalQueue.clear();
		}

		this.mountContext.generation = !generation;
		profiler().pop();
	}

	private void layoutRootElement(RootElement root, WidgetInstance<?> element) {
		element.updateMount(root);
		int scaledWindowWidth = this.scaledWindowWidth;
		int scaledWindowHeight = this.scaledWindowHeight;
		element.setAvailableSize(scaledWindowWidth, scaledWindowHeight);
		element.stencilLayout(this.stateContext);
		element.setPosition(0f, 0f);
		element.traceLayout();
		element.setClippedRect(0f, 0f, scaledWindowWidth, scaledWindowHeight);
		element.scissorLayout();
	}

	public boolean charTyped(int codePoint) {
		return iterateDown(CharTypedHandler.class, handler -> handler.charTyped(codePoint));
	}

	public boolean keyPressed(InputKey keyCode, int scanCode, InputMods modifiers) {
		return iterateDown(KeyPressHandler.class, handler -> handler.keyPressed(keyCode, scanCode, modifiers));
	}

	public boolean keyReleased(InputKey keyCode, int scanCode, InputMods modifiers) {
		return iterateDown(KeyReleaseHandler.class, handler -> handler.keyReleased(keyCode, scanCode, modifiers));
	}

	public boolean mousePressed(double mouseX, double mouseY, int button) {
		return iterateDown(MousePressHandler.class, handler -> handler.mousePressed(mouseX, mouseY, button));
	}

	public boolean mouseReleased(double mouseX, double mouseY, int button) {
		return iterateDown(MouseReleaseHandler.class, handler -> handler.mouseReleased(mouseX, mouseY, button));
	}

	public boolean mouseMoved(double mouseX, double mouseY) {
		this.mouseX = (float) mouseX;
		this.mouseY = (float) mouseY;
		return iterateDown(element -> element.mouseMoveHandler, handler -> handler.mouseMoved(mouseX, mouseY));
	}

	public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
		return iterateDown(element -> element.mouseScrollHandler,
				handler -> handler.mouseScrolled(mouseX, mouseY, delta));
	}

	public boolean mouseDragged(double mouseX, double mouseY, int button, double deltaX, double deltaY) {
		return iterateDown(element -> element.mouseDragHandler,
				handler -> handler.mouseDragged(mouseX, mouseY, button, deltaX, deltaY));
	}

	public float lastMouseX() {
		return this.mouseX;
	}

	public float lastMouseY() {
		return this.mouseY;
	}

	/**
	 * Bottom-up element iteration.
	 */
	private <T> boolean iterateUp(Class<T> clazz, Predicate<T> consumer) {
		var orderedBuffer = this.orderedBuffer;
		var a = orderedBuffer.elements();
		var upr = orderedBuffer.size();
		if (upr > a.length) throw new AssertionError();
		for (int i = 0; i < upr;) {
			var element = a[i++];
			if (clazz.isInstance(element)) {
				if (consumer.test((T)element)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Top-down element iteration.
	 */
	private <T> boolean iterateDown(Function<WidgetInstance<?>, @Nullable T> filter, Predicate<T> consumer) {
		return this.orderedBuffer.iterateDown(filter, consumer);
	}

	@Deprecated
	private <T> boolean iterateDown(Class<T> clazz, Predicate<T> consumer) {
		return this.orderedBuffer.iterateDown(clazz, consumer);
	}

	public int renderBufferByteCount() {
		return this.canvas.renderBufferByteCount();
	}

	final class StateContextImpl implements StateContext {
		private final NonNullLong2WidgetInstanceHashMap elements = GuiContext.this.elements;

		private static long computeKey(long parentKey, Key.CallSite callSite, Widget widget) {
			profiler().push("computeKey");
			Hasher hasher = callSite instanceof Key.Global ? new Hasher() : new Hasher(parentKey);
			callSite.hash(hasher);
			hasher.trackIdentity(widget.getClass());
			profiler().pop();
			return hasher.key();
		}

		public <W extends Widget> WidgetInstance<W> getState(long parentKey, Key.CallSite callSite, W widget) {
			profiler().push("GuiContext.getState");
			long key;
			try {
				Objects.requireNonNull(callSite, "Call-site must be non-null");
				Objects.requireNonNull(widget, "Cannot get state of a null widget");
				key = computeKey(parentKey, callSite, widget);
			} catch (Throwable e) {
				throw GuiException.wrap(FMT."Computing key of \{widget}@??? under @%016x\{parentKey}", e);
			}

			WidgetInstance<W> element;

			try {
				if (Config.SANITIZE_KEYS) {
					//noinspection unchecked
					element = (WidgetInstance<W>) this.elements.computeIfAbsent(key, i -> {
						// TODO: confirm that WidgetInstantiationArgs is always either allocated in a thread-local buffer
						//  or not allocated at all.

						var e = Objects.requireNonNull(widget.instantiate(new WidgetInstantiationArgs(GuiContext.this, widget.getClass(), i, callSite)));
						Logger.log(() -> FMT."instantiated element: \{widget}@%016x\{i} -> \{e}", List.of(DEBUG));
						return e;
					});
				} else {
					//noinspection unchecked
					element = (WidgetInstance<W>) this.elements.computeIfAbsent(key, i -> {
						// TODO: confirm that WidgetInstantiationArgs is always either allocated in a thread-local buffer
						//  or not allocated at all.

						var e = Objects.requireNonNull(widget.instantiate(new WidgetInstantiationArgs(GuiContext.this, widget.getClass(), i, null)));
						Logger.log(() -> FMT."instantiated element: \{widget}@%016x\{i} -> \{e}", List.of(DEBUG));
						return e;
					});
				}
//				{
//					element = Objects.requireNonNull(widget.instantiate(new WidgetInstantiationArgs(GuiContext.this, widget.getClass(), key)));
//					Logger.log(() -> FMT."instantiated element: \{widget}@%016x\{key} -> \{element}", List.of(DEBUG));
//				}
				checkForKeyCollision(callSite, widget, element);

				profiler().push("element.update");
				var initialized = element.initialized;
				element.configure(this, widget, !initialized);
				if (!initialized) element.initialized = true;
				var element_ = element;
				Logger.log(() -> FMT."updated element: \{widget}@%016x\{key} -> \{element_}", List.of(DEBUG));
				profiler().pop();

				profiler().pop();
				return element;
			} catch (Throwable e) {
				throw GuiException.wrap(FMT."Getting state of \{widget}@%016x\{key}", e);
			}
		}

		@Override
		public <W extends Widget> WidgetInstance<W> getState(WidgetInstance<?> parent, Key.CallSite callSite, W widget) {
			return getState(parent.key, callSite, widget);
		}
	}

	private static <W extends Widget> void checkForKeyCollision(Key.CallSite callSite, W widget, WidgetInstance<W> element) {
		if (Config.CHECKS && element.widgetClass != widget.getClass()) {
			Logger.log(() -> STR."""
			Key collision (widget class): \{element.widgetClass.getName()} != \{widget.getClass().getName()}
			\{element}
			\{widget}""");
		}
		if (Config.CHECKS && Config.SANITIZE_KEYS && !callSite.equals(element.keyObject)) {
			Logger.log(() -> STR."""
			Key collision (call-site): \{element.keyObject} != \{callSite}
			\{element}
			\{widget}""");
		}
	}

	void ensureElementIsManagedOnMount(WidgetInstance<?> element) {
		this.elements.putIfAbsent(element.key, element);
	}

	static final class MountContextImpl implements MountContext {
		private static final List<String> TRACE_TAGS = List.of(DEBUG, "MOUNT_OP");

		private final CanvasImpl canvas;
		private final WidgetInstanceList orderedBuffer;
		private final ShortArrayList treeBuffer;

		private final ScissorStack scissorStack = new ScissorStack();
		private float scissorTopMinX = Float.NaN;
		private float scissorTopMinY = Float.NaN;
		private float scissorTopMaxX = Float.NaN;
		private float scissorTopMaxY = Float.NaN;

		long time;
		boolean generation;
		private short treeDepth = 0;

		public MountContextImpl(GuiContext context) {
			this.canvas = context.canvas;
			this.orderedBuffer = context.orderedBuffer;
			this.treeBuffer = context.treeBuffer;
		}

		@Override
		public void pushScissor(float x, float y, float width, float height) {
			LOGGER.log(() -> STR."\{" ".repeat(Short.toUnsignedInt(this.treeDepth))}MountContext.pushScissor(\{x}, \{y}, \{width}, \{height})", TRACE_TAGS);
			this.canvas.internalPushScissor(x, y, width, height);
			var maxX = x + width;
			var maxY = y + height;
			this.scissorTopMinX = x;
			this.scissorTopMinY = y;
			this.scissorTopMaxX = maxX;
			this.scissorTopMaxY = maxY;
			this.scissorStack.push(x, y, maxX, maxY);
		}

		@Override
		public void popScissor() {
			LOGGER.log(() -> STR."\{" ".repeat(Short.toUnsignedInt(this.treeDepth))}MountContext.popScissor()", TRACE_TAGS);
			this.canvas.internalPopScissor();
			var scissorStack = this.scissorStack;
			scissorStack.pop();
			if (scissorStack.isEmpty()) {
				this.scissorTopMinX = Float.NaN;
				this.scissorTopMinY = Float.NaN;
				this.scissorTopMaxX = Float.NaN;
				this.scissorTopMaxY = Float.NaN;
			} else {
				this.scissorTopMinX = scissorStack.topMinX();
				this.scissorTopMinY = scissorStack.topMinY();
				this.scissorTopMaxX = scissorStack.topMaxX();
				this.scissorTopMaxY = scissorStack.topMaxY();
			}
		}

		private boolean shouldCull(WidgetInstance<?> element) {
			float x = element.allocatedX();
			float y = element.allocatedY();
			float width = element.computedWidth();
			float height = element.computedHeight();
			var scissorMinX = this.scissorTopMinX;
			var scissorMinY = this.scissorTopMinY;
			var scissorMaxX = this.scissorTopMaxX;
			var scissorMaxY = this.scissorTopMaxY;
			// comparison is false if either side is NaN
			// OR them all to find out if any are outside
			var shouldCull = scissorMinX > x + width |
							 scissorMinY > y + height |
							 x > scissorMaxX |
							 y > scissorMaxY;
//			element.lastShouldPaint = !shouldCull;
			return shouldCull;
		}

		private short mountCommon0(WidgetInstance<?> element) {
			this.orderedBuffer.add(element);
			short depth = this.treeDepth++;
			this.treeBuffer.add(depth);
			return depth;
		}

		@Override
		public void iSpecMount(WidgetInstance<?> element) {
			assert element != null : "Attempted to mount a null element";

			try {
				element.generation = this.generation;

//				if (shouldCull(element)) {
//					// FIXME: because this skips calling element.mount(), we can end up with elements being disposed
//					//  that are still references in other elements. That wouldn't be an issue, except that those if
//					//  those elements become visible again, they will no longer be remembered by the GuiContext, and
//					//  will never be disposed again.
////					LOGGER.log(Level.INFO, "culling");
//					return;
//				}

				if (element.painter != null) {
					profiler().push("paint");
					element.painter.paint(this.time, this.canvas);
					profiler().pop();
				}

				var depth = this.mountCommon0(element);

				element.mount(this);

				assert this.treeDepth == depth + 1;
				this.treeDepth = depth;

				if (element.postPainter != null) {
					profiler().push("paint");
					element.postPainter.paintPost(this.time, this.canvas);
					profiler().pop();
				}

				// null-checking here with a cache field instead of instanceof-checking for better performance.
				if (element.debugLayer != null) {
					assert element.debugRect != null;
					profiler().push("paint");
					if (this.canvas.isDebugLayerEnabled(element.debugLayer)) {
						element.debugRect.updatePositionAndSize(element.allocatedX(),
								element.allocatedY(),
								element.computedWidth(),
								element.computedHeight());
						this.canvas.upload(element.debugRect);
					}
					profiler().pop();
				}
			} catch (Throwable e) {
				throw GuiException.wrap(FMT."Mounting \{element.toStringShort()}\n\{element}", e);
			}
		}

		@Override
		public void mount(WidgetInstance<?> element) {
			LOGGER.log(() -> STR."\{" ".repeat(Short.toUnsignedInt(this.treeDepth))}MountContext.mount(\{element.toStringShort()})", TRACE_TAGS);
			element.uSpecMount(this);
		}
	}
}
