package dev.pfaff.altargui.render;

import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.internal.util.GrowableMemorySegment;
import dev.pfaff.altargui.shader.BuiltInstanceLayout;
import dev.pfaff.altargui.shader.Program;
import dev.pfaff.altargui.shader.Programs;
import dev.pfaff.altargui.shader.ShaderProg2D;
import dev.pfaff.glass.opengl.enums.GLBlendEquationModeEXT;
import dev.pfaff.glass.opengl.enums.GLBlendingFactor;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;
import dev.pfaff.glass.opengl.enums.GLEnableCap;
import dev.pfaff.log4truth.NamedLogger;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SequenceLayout;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.util.List;

import static dev.pfaff.altargui.CanvasImpl.DRAW_TAGS;
import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.internal.Config.GL_CHECKS;
import static dev.pfaff.altargui.internal.Config.UNSAFE_OPTIMIZATIONS;
import static dev.pfaff.altargui.util.MethodHandleUtil.buildMH;
import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.ValueLayout.JAVA_INT;
import static java.lang.foreign.ValueLayout.JAVA_LONG;
import static java.lang.invoke.MethodHandles.*;
import static java.lang.invoke.MethodType.methodType;

public abstract class LayeredRenderBuffer implements AutoCloseable {
	private static final NamedLogger LOGGER = NamedLogger.forCaller();

	private static final SequenceLayout ENTRY_LAYOUT = sequenceLayout(4, JAVA_INT);

	protected final List<Program<ShaderProg2D>> programs = List.of(Programs.RECT_RENDER_LAYER,
			Programs.TEXTURED_RECT_RENDER_LAYER,
			Programs.TEXT_RENDER_LAYER);
	private static final MethodHandle MH_drawNext = buildMH(() -> {
		return lookup().findVirtual(LayeredRenderBuffer.class,
				"drawNext",
				methodType(void.class, int.class, Program.class, MemorySegment.class, long.class));
	});
	private final MethodHandle programDrawNextDispatch;
	// FIXME: need to rebuild on shader reload
	protected final List<BuiltInstanceLayout> instanceLayouts
			= this.programs.stream().map(layer -> layer.instanceLayout().build(layer.shader().get())).toList();

	/**
	 * Entries are stored as pairs of {@code (program, data1, data2, data3)}
	 */
	private MemorySegment entries = MemorySegment.NULL;
	private long entriesLen = 0L;
	protected final RawBuffer instancesBuffer;
	private int currentProgram = -1;
	private boolean uploading = false;

	public LayeredRenderBuffer() {
		this.instancesBuffer = RawBuffer.create(GLBufferTarget.ARRAY_BUFFER, true);
		var MH_drawNext = LayeredRenderBuffer.MH_drawNext.bindTo(this);
		this.programDrawNextDispatch = MethodHandles.tableSwitch(
				GL_CHECKS ? dropArguments(MethodHandles.throwException(void.class, IndexOutOfBoundsException.class)
							.bindTo(new IndexOutOfBoundsException()),
							0,
							int.class,
							MemorySegment.class,
							long.class)
						: MethodHandles.empty(methodType(void.class, int.class, MemorySegment.class, long.class)),
				this.programs.stream()
						.map(prog -> insertArguments(MH_drawNext, 1, prog))
						.toArray(MethodHandle[]::new));
	}

	public final int currentProgram() {
		return this.currentProgram;
	}

	public final int byteCount() {
		return this.instancesBuffer.length();
	}

	protected abstract void onRecreate();

	/**
	 * @param capacity the maximum number of bytes that will be uploaded.
	 */
	public void beginUploading(int capacity) {
		if (GL_CHECKS) {
			if (this.uploading) throw new IllegalStateException("Already uploading!");
			this.uploading = true;
		}
		this.entriesLen = 0L;
		this.instancesBuffer.reset(capacity, this::onRecreate);
		this.instancesBuffer.beginUploading();
	}

	public final void finishUploading() {
		if (GL_CHECKS) {
			if (!this.uploading) throw new IllegalStateException("Not uploading!");
			this.uploading = false;
		}
		this.endCurrentLayer();
		this.instancesBuffer.finishUploading();
	}

	@Override
	public void close() {
		this.instancesBuffer.close();
	}

	/**
	 * @param ms a buffer of 3 integers. Store whatever you need here.
	 * @param programId the shader program used for the layer.
	 * @param start the offset in units of instances of the start of the program's vertex data.
	 * @param end the offset in units of instances of the end of the program's vertex data.
	 */
	protected abstract void putDataForEntry(MemorySegment ms, int programId, int start, int end);

	/**
	 * @param ms a buffer of 3 integers.
	 * @param programId
	 */
	protected abstract VAO getVAOForEntry(MemorySegment ms, int programId);

	/**
	 * The VAO will be bound by the caller.
	 *
	 * @param ms a buffer of 3 integers.
	 */
	protected abstract void drawEntry(MemorySegment ms, int programId, Program<?> program);

	private void useProgram(int program) {
		if (this.currentProgram != program) {
			this.endCurrentLayer();
			this.currentProgram = program;
			var stride = this.instanceLayouts.get(program).stride();
			RawBuffer instancesBuffer = this.instancesBuffer;
			var len = instancesBuffer.length();
			var index = (len + (stride - 1)) / stride;
			var alignedLen = index * stride;
			instancesBuffer.setLength(instancesBuffer.length() + (alignedLen - len));

			var minCapacity = this.entriesLen + Integer.BYTES * 4;
			if (this.entries.byteSize() < minCapacity) {
				this.entries = GrowableMemorySegment.growAlways(this.entries, minCapacity, 16);
			}
			this.entries.set(JAVA_LONG, this.entriesLen, ((long)program) | (((long)index) << 32));
			this.entriesLen += Long.BYTES;

//			LOGGER.log(Level.INFO, () -> STR."began program \{program} at \{len} aligned up to \{alignedLen} (index \{index})");
		}
	}

	public final boolean endCurrentLayer() {
		int program = this.currentProgram;
		if (program == -1) return false;
		int index = this.instancesBuffer.length() / this.instanceLayouts.get(program).stride();
		int startIndex = this.entries.get(JAVA_INT, this.entriesLen - Integer.BYTES);
		this.entriesLen += Long.BYTES;
		this.putDataForEntry(
				this.entries.asSlice(this.entriesLen - (Integer.BYTES * 3),Integer.BYTES * 3),
				program,
				startIndex,
				index);
//			LOGGER.log(Level.INFO, () -> STR."ended program \{program} at \{this.instancesBuffer.length()} (index \{index})");
		this.currentProgram = -1;
		return true;
	}

	public final void iSpecUpload(RenderObject object, int program) {
//		LOGGER.log(() -> STR."uploading \{object}");
		this.useProgram(program);
		this.instancesBuffer.upload(object);
		if (GL_CHECKS) {
			var layout = this.instanceLayouts.get(program);
			var stride = layout.stride();
			var trailing = this.instancesBuffer.length() % stride;
			if (trailing != 0) {
				throw new BufferOverflowException(STR."""
				Upload of \{object} left \{trailing} trailing bytes in buffer.
				  This is either an underflow or an overflow.
				  The object consists of \{object.instanceCount()} instances (\{object.size()} bytes)
				  The object layout is \{object.instanceLayout()} (\{object.instanceLayout().byteSize()} bytes)
				  The layer has a stride of \{stride} bytes
				  The layer id is \{program}
				  The layer layout is \{layout}""");
			}
		}
	}

	public final void upload(RenderObject object) {
		object.uSpecUploadLayered(this);
	}

	public final class Drawer implements AutoCloseable {
		private final MemorySegment entries;
		private long offset;

		private Drawer(MemorySegment entries) {
			this.entries = entries;
		}

		/**
		 * Draws the next layer.
		 */
		public boolean drawNext() {
			if (this.offset >= this.entries.byteSize()) {
				return false;
			}
			int programId = this.entries.get(JAVA_INT, this.offset);
			try {
				LayeredRenderBuffer.this.programDrawNextDispatch.invokeExact(programId, this.entries, this.offset);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
			this.offset += ENTRY_LAYOUT.byteSize();
			return true;
		}

		public void drawRemaining() {
			while (drawNext()) {}
		}

		@Override
		public void close() {
			if (Config.GL_ALWAYS_UNBIND) GLUtil.glxUnbindVertexArray();
		}
	}

	public final Drawer drawer() {
		GL.glEnable(GLEnableCap.BLEND);
		GL.glBlendEquation(GLBlendEquationModeEXT.FUNC_ADD);
		GL.glBlendFunc(GLBlendingFactor.SRC_ALPHA, GLBlendingFactor.ONE_MINUS_SRC_ALPHA);
		GL.glDisable(GLEnableCap.DEPTH_TEST);

		var buf = this.entries.asSlice(0, this.entriesLen);
		if (buf.byteSize() % (Integer.BYTES * 4) != 0) throw new IllegalStateException();
		return new Drawer(buf);
	}

	private void drawNext(int programId, Program<ShaderProg2D> program, MemorySegment entries, long offset) {
		MemorySegment entry;
		if (UNSAFE_OPTIMIZATIONS) {
			entry = entries.reinterpret(offset + ENTRY_LAYOUT.byteSize(), Arena.global(), _ -> {})
					.asSlice(offset, ENTRY_LAYOUT);
		} else {
			entry = entries.asSlice(offset, ENTRY_LAYOUT);
		}

		var shader = program.shader().get();

		LOGGER.log(() -> STR."draw(shader=\{shader.name()})", DRAW_TAGS);

		var entryData = entry.asSlice(Integer.BYTES);
		if (UNSAFE_OPTIMIZATIONS) {
			entryData = entryData.reinterpret(Integer.BYTES * 3);
		}

		var vao = this.getVAOForEntry(entryData, programId);
		vao.bind();

		program.drawSetup().accept(shader);
		GL.checkError();

		shader.bind();
		GL.checkError();

		this.drawEntry(entryData, programId, program);
	}
}
