package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.altargui.io.IOUtil;
import dev.pfaff.altargui.libimage.*;
import dev.pfaff.altargui.util.BasicFuture;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.altargui.util.ThreadFuture;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.MutPointer;
import dev.pfaff.glass.opengl.enums.*;
import dev.pfaff.log4truth.NamedLogger;
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import org.jspecify.annotations.Nullable;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.ref.SoftReference;
import java.nio.file.Path;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Phaser;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.render.GLContextPool.GL_CONTEXT_POOL;
import static dev.pfaff.altargui.render.GLContextPool.POOL_ACQUIRED_GL;
import static dev.pfaff.altargui.util.BasicFuture.completedTask;
import static dev.pfaff.altargui.util.BasicFuture.executeOn;
import static dev.pfaff.altargui.util.ThreadFuture.spawnVirtualTask;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.ERROR;
import static dev.pfaff.log4truth.format.Format.STANDARD_DURATION;
import static java.lang.foreign.ValueLayout.JAVA_BYTE;
import static java.lang.foreign.ValueLayout.JAVA_INT;

/**
 * <p>
 * <code>
 * Arrays: | 1  | 2  | 3  |<br/>
 * Active: | D  | U* | U  |
 * </code>
 * <p>
 * States:
 * <ol>
 *     <li><strong>D</strong> clean for drawing</li>
 *     <li><strong>U</strong> accepting uploads</li>
 *     <li><strong>U*</strong> finishing upload</li>
 * </ol>
 * <p>
 * When a buffer with state {@code U*} is finished, it transitions to state {@code D}, the previous buffer
 * transitions to state {@code U}, and the next buffer transitions to state {@code U*}. Thanks to this model, the
 * states of all three buffers may be represented using a single 2-bit index.
 * <p>
 *
 * <code>
 * |---|---|---|<br/>
 * | A | B | C |<br/>
 * |---|---|---|
 * </code>
 * <p>
 * Initial state:
 * <ul>
 *     <li><strong>A)</strong> Drawing</li>
 *     <li><strong>B)</strong> Finishing (nothing)</li>
 *     <li><strong>C)</strong> Uploading</li>
 * </ul>
 *
 * <p>
 * Next state, triggered on job submission to C:
 * <ul>
 *     <li><strong>A)</strong> Uploading</li>
 *     <li><strong>B)</strong> Drawing</li>
 *     <li><strong>C)</strong> Finishing (at least 1 job)</li>
 * </ul>
 *
 * <p>
 * Next state, triggered on job completion in C:
 * <ul>
 *     <li><strong>A)</strong> Finishing (possibly nothing)</li>
 *     <li><strong>B)</strong> Uploading</li>
 *     <li><strong>C)</strong> Drawing</li>
 * </ul>
 *
 * <p>
 * Next state, triggered on job completion in A (if any) and job submission to B:
 * <ul>
 *     <li><strong>A)</strong> Drawing</li>
 *     <li><strong>B)</strong> Finishing (possibly nothing)</li>
 *     <li><strong>C)</strong> Uploading</li>
 * </ul>
 */
public final class TextureAllocator implements AutoCloseable {
	private static final NamedLogger LOGGER = new NamedLogger("TexAlloc");

	public static final byte TBS_COUNT = 3;

	public static final int MAX_TEXTURE_SIZE;

	static {
		// cannot use lambdas across threads in static initializer
		MAX_TEXTURE_SIZE = executeOn(new Callable<Integer>() {
			@Override
			public Integer call() throws Exception {
				return GL.glGetInteger(GLGetPName.MAX_TEXTURE_SIZE);
			}
		}, GLThread::submitToGLThread);
		if (MAX_TEXTURE_SIZE == 0) {
			throw new IllegalStateException("MAX_TEXTURE_SIZE must be greater than 0");
		}
	}

	/**
	 * The minimum allocation size in pixels. Herein we often measure in "units" rather than pixels for efficiency.
	 */
	public static final int UNIT = Math.divideExact(MAX_TEXTURE_SIZE, 256);

	static {
		if (Integer.bitCount(UNIT) != 1) {
			throw new UnsupportedOperationException(STR."Only power-of-two unit sizes are supported, but \{UNIT} was calculated");
		}
	}

	private static final AtomicInteger UPLOAD_ID = new AtomicInteger(0);

	/**
	 * @param rotation the rotation of the active texture for drawing from the zeroth.
	 */
	private record TBS(byte rotation, TBS1 completing, TBS1 initiating) {
		public TBS() {
			this((byte) 0, new TBS1(), new TBS1());
		}

		@Nullable
		public TBS pollNext() {
			if (this.completing.dirty.get() || this.initiating.dirty.get()) {
				if (this.completing.activeCounter.compareAndSet(0, Integer.MAX_VALUE)) {
					return new TBS((byte) ((byte) (this.rotation + 1) % 3), this.initiating, new TBS1());
				}
			}
			return null;
		}
	}

	/**
	 * <p>
	 * registering:
	 * MAX_VALUE + 1 -> MIN_VALUE
	 * => signals that we cannot add
	 * <p>
	 * completing:
	 * 1 - 1 -> 0
	 * <p>
	 * checking:
	 * compareAndSet(0, Integer.MAX_VALUE)
	 * => if true, all are complete, and no more will be added.
	 *
	 * @param dirty whether the buffer is being or has been modified
	 * @param activeCounter the number of active uploads to the buffer
	 * @param onAvailableSync triggered when the previous batch (another {@link TBS1}) is complete and the next batch
	 *                        has been made available.
	 */
	private record TBS1(AtomicBoolean dirty, AtomicInteger activeCounter, Phaser onAvailableSync) {
		public TBS1() {
			this(new AtomicBoolean(), new AtomicInteger(), new Phaser(1));
		}
	}

	/**
	 * The size (in {@link #UNIT}s) of each layer in the allocator.
	 */
	private final int sizeUnits;
	private int layers;
	private final AtomicReference<TBS> tbs = new AtomicReference<>(new TBS());
	private final TextureArray textureArrayA = new TextureArray();
	private final TextureArray textureArrayB = new TextureArray();
	private final TextureArray textureArrayC = new TextureArray();
	private final int textureBufferId;
	private final BitSet allocations = new BitSet();
	private final Object2ObjectOpenHashMap<LoadableTexture, Object> cache =
			new Object2ObjectOpenHashMap<>();
	private final Semaphore lock = new Semaphore(1);

	/**
	 * @param size the width and height of each layer.
	 */
	public TextureAllocator(int size, int initialLayers) {
		if (size > MAX_TEXTURE_SIZE) {
			throw new IllegalArgumentException("Layer size too large");
		}
		this.sizeUnits = size / UNIT;
		if (size % UNIT != 0) throw new IllegalArgumentException(STR."Invalid layer size: \{size}");
		this.layers = initialLayers;
		this.textureBufferId = GL.glGenBuffer();
	}

	public void reset() {
		for (byte i = 0; i < TBS_COUNT; i++) {
			this.reset(this.getArray(i));
		}
	}

	private void reset(TextureArray textureArray) {
		var size = this.sizeUnits * UNIT;

		try (var _ = textureArray.bound()) {
			textureArray.allocate(size, size, this.layers, TextureOptions.DEFAULT);
		}
	}

	private TextureArray getArray(byte rotation) {
		if (rotation < 0 || rotation > TBS_COUNT) throw new AssertionError();
		return switch (tbBufIndex(rotation)) {
			case 0 -> this.textureArrayA;
			case 1 -> this.textureArrayB;
			case 2 -> this.textureArrayC;
			default -> throw new AssertionError();
		};
	}

	private static int tbBufIndex(byte rotation) {
		return rotation % TBS_COUNT;
	}

	public InfallibleAutoClosable boundForDraw() {
		var tbs = this.tbs.get();
		var nextTbs = tbs.pollNext();
		if (nextTbs != null) {
			{
				var tbs_ = tbs;
				LOGGER.log(() -> STR."Rotating TBS:\n\t\{tbs_}\n\t\{nextTbs}", List.of(DEBUG));
			}
			long start = System.nanoTime();
			tbs = nextTbs;
			this.tbs.set(tbs);
//			tbs.completing.nextAvailable.countDown();
			// Two-step process. First announce that the next is available, then wait for other threads to submit their
			// jobs to it.
			// TODO: don't block the main thread. Instead, arrive and then check if others have arrived on each call to
			//  this method.
			tbs.completing.onAvailableSync.arriveAndAwaitAdvance();
			tbs.completing.onAvailableSync.arriveAndAwaitAdvance();
			tbs.completing.onAvailableSync.forceTermination();
			long elapsed = System.nanoTime() - start;
			LOGGER.log(() -> STR."Rotated TBS in \{STANDARD_DURATION.nanosToString(elapsed)}", List.of(DEBUG));
		}
		return this.getArray(tbs.rotation).bound();
	}

	/**
	 * <p>
	 * Gets a shader to the given texture. If not loaded, the texture will be loaded and a slot allocated for it.
	 * <p>
	 * This method must only be called from the main thread.
	 */
	public BasicFuture<TextureHandle> alloc(LoadableTexture key) {
		{
			var cacheValue = this.cache.get(key);
			// this inspection is wrong
			//noinspection ConstantValue
			if (cacheValue != null) {
				var handle = tryDerefCacheValue(cacheValue);
				if (handle != null) return handle;
			}
		}
		BasicFuture<TextureHandle> future = spawnVirtualTask(() -> {
			ImageData image;
			try {
				image = key.load();
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture load failed", t, List.of(ERROR));
				return Objects.requireNonNull(tryDerefCacheValueJoin(this.cache.get(SinglePixelLoadableTexture.WHITE)));
			}

			try {
				// allocate a new image in the texture array.
				var handle = this.allocNew(key, image);

				// try getting from the cache again before storing to avoid keeping alive duplicate allocations.
				executeOn(() -> {
					assert this.cache.get(key) instanceof ThreadFuture<?>;
					this.cache.put(key, new SoftReference<>(handle));
				}, GLThread::submitToGLThread);
				return handle;
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture allocation failed", t, List.of(ERROR));
				return Objects.requireNonNull(tryDerefCacheValueJoin(this.cache.get(SinglePixelLoadableTexture.WHITE)));
			}
		});
		this.cache.put(key, future);
		return future;
	}

	private static @Nullable BasicFuture<TextureHandle> tryDerefCacheValue(Object cacheValue) {
		if (cacheValue instanceof SoftReference<?> ref) {
			var handle = (TextureHandle) ref.get();
			if (handle != null) return completedTask(handle);
		} else {
			//noinspection unchecked
			return (BasicFuture<TextureHandle>) cacheValue;
		}
		return null;
	}

	private static @Nullable TextureHandle tryDerefCacheValueJoin(Object cacheValue) {
		if (cacheValue instanceof SoftReference<?> ref) {
			return (TextureHandle) ref.get();
		} else {
			//noinspection unchecked
			return ((BasicFuture<TextureHandle>) cacheValue).join();
		}
	}

	private TextureHandle allocNew(LoadableTexture key, ImageData image) throws InterruptedException {
		int width = image.info().width();
		int height = image.info().height();

		// align size up to the next multiple of UNIT.
		// these are in units, not pixels
		int unitWidth = (width + (UNIT - 1)) / UNIT;
		int unitHeight = (height + (UNIT - 1)) / UNIT;
		assert unitWidth * UNIT >= width && unitHeight * UNIT >= height : STR."width and/or height was improperly aligned to unit size:\n\t\{width} > \{unitWidth * UNIT}\n\tor\n\t\{height} > \{unitHeight * UNIT}";

		int sizeUnits = this.sizeUnits;

		if (unitWidth > sizeUnits || unitHeight > sizeUnits) {
			throw new IllegalArgumentException(STR."Size of image exceeds size of allocator layers: \{width} || \{height} > \{this.sizeUnits * UNIT}");
		}

		LOGGER.log(() -> STR."Searching for a free region to allocate \{key} an image of \{width}x\{height}", List.of(DEBUG));

		DetailedHandle detailed;

		this.lock.acquire();
		try {
			// find a free region
			int allocationsSize = sizeUnits * sizeUnits * this.layers;
			int fromSearchUpr = allocationsSize - unitWidth - (unitHeight - 1) * sizeUnits;
			int layers = this.layers;
			int sizeUnitsSq = sizeUnits * sizeUnits;

			int xUpr = sizeUnits - unitHeight;

			int from = -1, iUpr = -1;

			search:
			for (int layerStart = 0; layerStart < layers * sizeUnitsSq; layerStart += sizeUnitsSq) {
				from = layerStart;
				int layerUpr = from + sizeUnitsSq;
				while (from <= fromSearchUpr) {
					assert from % sizeUnits == 0 :
							STR."not at the start of a row: from=\{from}";
					int rowUpr = from + xUpr;
					for (; from < rowUpr; from++) {
						iUpr = from + unitHeight * sizeUnits;
						assert iUpr <= layerUpr;

						if (isRegionFree(from, iUpr, unitWidth)) break search;
					}
					from += unitHeight;
				}
				iUpr = -1;
			}

			if (iUpr == -1) {
				throw new UnsupportedOperationException("TODO: realloc");
			}
			assert from != -1;

			detailed = this.makeHandle(from, iUpr, unitWidth, width, height);
		} finally {
			this.lock.release();
		}

		upload(image, detailed, width, height);
		return detailed.handle();
	}

	private record ImageSrc(MemorySegment pixels, GLPixelFormat format, GLPixelType type) {}

	private static ImageSrc getImageSrc(ImageData image) {
		return switch (image.info().colorType()) {
			case ColorType.Rgba8 -> new ImageSrc(swizzleRgba8ToBgra8(image.pixels()), GLPixelFormat.BGRA, GLPixelType.UNSIGNED_INT_8_8_8_8_REV);
			case ColorType.Rgb8 -> new ImageSrc(swizzleRgb8ToBgra8(image.pixels()), GLPixelFormat.BGRA, GLPixelType.UNSIGNED_INT_8_8_8_8_REV);
			default -> new ImageSrc(image.pixels(), image.info().colorType().glPixelFormat(), image.info().colorType().glChannelType());
		};
	}

	private void upload(ImageData image, DetailedHandle detailed, int width, int height) {
		int id = UPLOAD_ID.getAndAdd(1);

		var src = getImageSrc(image);

		var handle = detailed.handle();
		LOGGER.log(() -> STR."[Upload #\{id}] Allocated texture handle \{handle.layer}:\{handle.u},\{handle.v} for image of \{width}x\{height} at \{
				detailed.u() * UNIT},\{detailed.v() * UNIT}", List.of(DEBUG));
		var completer = new CompletableFuture<@Nullable Void>();
		GL_CONTEXT_POOL.acquireExecutor().execute(() -> {
			try {
				var gl = POOL_ACQUIRED_GL.get();
				assert gl == GL;

				var bufId = GL.glGenBuffer();
				try {
					GL.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, bufId);
					try {
						GL.glBufferData(GLBufferTarget.PIXEL_UNPACK_BUFFER,
								src.pixels().byteSize(),
								new ConstPointer<>(src.pixels()),
								GLBufferUsage.STREAM_DRAW);

						TBS tbs;
						while (true) {
							tbs = this.tbs.get();
							if (tbs.initiating.activeCounter.addAndGet(1) > 0) break;
							// undo
							tbs.initiating.activeCounter.getAndAdd(-1);
						}
						for (int i = 0; i < TBS_COUNT;) {
							long start = System.nanoTime();
							tbs.initiating.dirty.set(true);

							var array = this.getArray((byte) ((byte) (tbs.rotation + 2) % TBS_COUNT));

							// upload
							try {
								try (var _ = array.bound()) {
									array.startUploading();
									array.uploadFromPixelUnpackBuffer(handle.layer,
											detailed.u() * UNIT,
											detailed.v() * UNIT,
											width,
											height,
											src.format(),
											src.type());
								}

								int syncId = GL.glFenceSync(GLSyncCondition.SYNC_GPU_COMMANDS_COMPLETE, 0);
								try {
									GL.glClientWaitSync(syncId, GLSyncObjectMask.SYNC_FLUSH_COMMANDS_BIT.glEnum(), Long.MAX_VALUE);
								} finally {
									GL.glDeleteSync(syncId);
								}
							} finally {
								var oldTbs = tbs;
								long elapsed = System.nanoTime() - start;
								LOGGER.log(() -> STR."[Upload #\{id}] Uploaded to uploadBuf=#\{tbBufIndex((byte) (oldTbs.rotation + 2))} (drawBuf=#\{tbBufIndex(oldTbs.rotation)}) activeUploads=\{oldTbs.initiating.activeCounter.get()} (self included)\n\ttook \{STANDARD_DURATION.nanosToString(elapsed)}", List.of(DEBUG));

								oldTbs.initiating.onAvailableSync.register();

								// We're finished, so decrement the active count.
								// It is important that we do this only *after* registering to onAvailableSync to avoid
								// race conditions.
								oldTbs.initiating.activeCounter.getAndAdd(-1);

								// wait for the next to be available
								oldTbs.initiating.onAvailableSync.arriveAndAwaitAdvance();

								var newTbs = this.tbs.get();
								assert newTbs.completing == oldTbs.initiating;
								tbs = newTbs;

								if (++i != TBS_COUNT) {
									LOGGER.log(() -> STR."[Upload #\{id}] Ready for next upload", List.of(DEBUG));

									var n = newTbs.initiating.activeCounter.addAndGet(1);
									assert n > 0;
								}

								oldTbs.initiating.onAvailableSync.arriveAndDeregister();

								if (i == 1) {
									// The first upload has just completed. Now, spawn a thread to await the rotation
									// to the texture, then notify the caller that the upload has completed. At that
									// point, to an external observer, everything *has* completed. However, the changes
									// will still need to be propagated to the remaining two buffers internally.
									Thread.startVirtualThread(() -> {
										LOGGER.log(() -> STR."[Upload #\{id}] Waiting for first upload to rotate into active", List.of(DEBUG));
										try {
											if (newTbs.initiating.onAvailableSync.register() >= 0) {
												LOGGER.log(() -> STR."[Upload #\{id}] Awaiting", List.of(DEBUG));
												newTbs.initiating.onAvailableSync.arriveAndAwaitAdvance();
												newTbs.initiating.onAvailableSync.arriveAndDeregister();
											}
											LOGGER.log(() -> STR."[Upload #\{id}] Complete", List.of(DEBUG));
											completer.complete(null);
										} catch (Throwable e) {
											completer.completeExceptionally(e);
										}
									});
								}
							}
						}
					} finally {
						GL.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, 0);
					}
				} finally {
					GL.glDeleteBuffer(bufId);
				}
			} catch (Throwable e) {
				completer.completeExceptionally(e);
			}
		});
		completer.join();
		LOGGER.log(() -> STR."[Upload #\{id}] Uploaded image to \{handle.layer}:\{handle.u},\{handle.v}", List.of(DEBUG));
	}

	private static MemorySegment swizzleRgba8ToBgra8(MemorySegment pixels) {
		if (pixels.byteSize() % 4 != 0) {
			throw new IllegalArgumentException("Invalid RGBA8 pixels");
		}
		var dst = Arena.ofAuto().allocate(pixels.byteSize(), 16);
		for (long i = 0; i < pixels.byteSize(); i += 4) {
			var pixel = pixels.get(JAVA_INT, i);
			// RGBA bytes -> ABGR int
			// ABGR int rotl 8 -> BGRA int
			// BGRA int rev -> ARGB int
			// Integer.reverseBytes(Integer.rotateLeft(pixel, 8))
			dst.set(JAVA_INT, i, (pixel & 0xff_00_ff_00) | ((pixel & 0xff) << 16) | ((pixel & 0xff_00_00) >>> 16));
		}
		return dst;
	}

	private static MemorySegment swizzleRgb8ToBgra8(MemorySegment pixels) {
		if (pixels.byteSize() % 3 != 0) {
			throw new IllegalArgumentException("Invalid RGB8 pixels");
		}
		var dst = Arena.ofAuto().allocate(pixels.byteSize() / 3 * 4, 16);
		for (long i = 0, j = 0; i < pixels.byteSize(); i += 3, j += 4) {
			var r = Byte.toUnsignedInt(pixels.get(JAVA_BYTE, i));
			var g = Byte.toUnsignedInt(pixels.get(JAVA_BYTE, i + 1));
			var b = Byte.toUnsignedInt(pixels.get(JAVA_BYTE, i + 2));
			dst.set(JAVA_INT, j, 0xff_00_00_00 | (r << 16) | (g << 8) | b);
		}
		return dst;
	}

	private DetailedHandle makeHandle(int from, int iUpr, int unitWidth, int textureWidth, int textureHeight) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// mark allocated
		for (int i = from; i < iUpr; i += sizeUnits) { // for each row
			// mark the region of each row allocated
			allocations.set(i, i + unitWidth);
		}

		var handle = MemorySegment.NULL.reinterpret(Arena.ofAuto(), _ -> {
			this.lock.acquireUninterruptibly();
			try {
				// mark free
				for (int i = from; i < iUpr; i += sizeUnits) {
					allocations.clear(i, i + unitWidth);
				}
			} finally {
				this.lock.release();
			}
		});

		int layer = from / (sizeUnits * sizeUnits);
		int u = from % sizeUnits;
		int v = (from / sizeUnits) % sizeUnits;
		return new DetailedHandle(handle, layer, sizeUnits, u, v, textureWidth, textureHeight);
	}

	private boolean isRegionFree(int from, int iUpr, int unitWidth) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// check if all are free
		for (int i = from; i < iUpr; i += sizeUnits) {
			for (int x = 0; x < unitWidth; x++) {
				if (allocations.get(i + x)) return false;
			}
		}
		return true;
	}

	@Override
	public void close() {
		this.textureArrayA.close();
		this.textureArrayB.close();
		GL.glDeleteBuffer(this.textureBufferId);
	}

	public void dump() {
		var arr = this.getArray(this.tbs.get().rotation);
		try {
			try (var _ = arr.bound()) {
				try (var arena = Arena.ofConfined()) {
					var segment = arena.allocate(JAVA_INT,
							(this.sizeUnits * UNIT) * (this.sizeUnits * UNIT * this.layers));
					GL.glGetTexImage(arr.target(),
							0,
							GLPixelFormat.RGBA,
							GLPixelType.UNSIGNED_BYTE,
							new MutPointer<>(segment));
					var encoded = LibImage.encodeImage(
							new ImageInfo(this.sizeUnits * UNIT, this.sizeUnits * UNIT * this.layers, ColorType.Rgba8),
							segment,
							ImageFormat.Png,
							EncoderProfile.Fast,
							arena);
					IOUtil.write(Path.of("allocated-texture.png"), encoded);
				}
			}

			try (var arena = Arena.ofConfined()) {
				var size = this.sizeUnits * this.sizeUnits * this.layers;
				var segment = arena.allocate(size);
				for (int i = 0; i < size; i++) {
					var pixel = this.allocations.get(i) ? 0xff : 0x00;
					segment.setAtIndex(JAVA_BYTE, i, (byte) pixel);
				}
				var encoded = LibImage.encodeImage(
						new ImageInfo(this.sizeUnits, this.sizeUnits * this.layers, ColorType.L8),
						segment,
						ImageFormat.Png,
						EncoderProfile.Fast,
						arena);
				IOUtil.write(Path.of("allocated-units.png"), encoded);
			}
		} catch (IOException e) {
			LOGGER.log(() -> "Failed to dump texture atlas", e, List.of(ERROR));
		}
	}

	/**
	 * @param u measured in {@link #UNIT}s
	 * @param v measured in {@link #UNIT}s
	 */
	record DetailedHandle(TextureHandle handle, int u, int v) {
		/**
		 * @param layer            index
		 * @param layerSizeInUnits the size of each layer, measured in {@link #UNIT}s
		 * @param u                measured in {@link #UNIT}s
		 * @param v                measured in {@link #UNIT}s
		 */
		public DetailedHandle(MemorySegment handle,
							  int layer, int layerSizeInUnits,
							  int u, int v,
							  int textureWidth, int textureHeight) {
			this(new TextureHandle(handle,
					layer,
					layerSizeInUnits * UNIT,
					u / (float) layerSizeInUnits,
					v / (float) layerSizeInUnits,
					textureWidth,
					textureHeight), u, v);
		}
	}
}
